@@ -12,6 +12,7 @@ def count_down_delay(delay):
         sleep(1)
     print("")
 
+
 DASH_SEPARATOR = 75 * "-"
 
 
@@ -23,7 +24,7 @@ def setup(*args, **kwargs):
         raise
     except SystemExit:
         raise
-    except:
+    except Exception:
         print(DASH_SEPARATOR)
         print("Sorry, your build failed. Try rerunning configure.py with "
                 "different options.")
@@ -212,7 +213,7 @@ def expand_value(v, options):
         for i in v:
             try:
                 exp_i = expand_value(i, options)
-            except:
+            except Exception:
                 pass
             else:
                 result.append(exp_i)
@@ -285,7 +286,7 @@ class ConfigSchema:
             if value is not None:
                 filevars[key] = value
 
-        keys = filevars.keys()
+        keys = list(filevars.keys())
         keys.sort()
 
         outf = open(filename, "w")
@@ -687,17 +688,20 @@ def substitute(substitutions, fname):
     fname_in = fname+".in"
     lines = open(fname_in, "r").readlines()
     new_lines = []
-    for l in lines:
+    for line in lines:
         made_change = True
         while made_change:
             made_change = False
-            match = var_re.search(l)
+            match = var_re.search(line)
             if match:
                 varname = match.group(1)
-                l = l[:match.start()] + str(substitutions[varname]) + l[match.end():]
+                line = (
+                        line[:match.start()]
+                        + str(substitutions[varname])
+                        + line[match.end():])
                 made_change = True
 
-            match = string_var_re.search(l)
+            match = string_var_re.search(line)
             if match:
                 varname = match.group(1)
                 subst = substitutions[varname]
@@ -706,9 +710,9 @@ def substitute(substitutions, fname):
                 else:
                     subst = '"%s"' % subst
 
-                l = l[:match.start()] + subst + l[match.end():]
+                line = line[:match.start()] + subst + line[match.end():]
                 made_change = True
-        new_lines.append(l)
+        new_lines.append(line)
     new_lines.insert(1, "# DO NOT EDIT THIS FILE -- "
             "it was generated by configure.py\n")
     new_lines.insert(2, "# %s\n" % (" ".join(sys.argv)))
@@ -823,3 +827,24 @@ def check_git_submodules():
                     print(DASH_SEPARATOR)
                     print("git submodules initialized successfully")
                     print(DASH_SEPARATOR)
+
+
+def check_pybind11():
+    try:
+        import pybind11  # noqa
+    except ImportError:
+        print(DASH_SEPARATOR)
+        print("Pybind11 is not installed.")
+        print(DASH_SEPARATOR)
+        print("Very likely, the build process after this message will fail.")
+        print("")
+        print("Simply press Ctrl+C and type")
+        print("python -m pip install pybind11")
+        print("to fix this. If you don't, the build will continue ")
+        print("in a few seconds.")
+        print("")
+        print("[1] https://pybind11.readthedocs.io/en/stable/")
+        print(DASH_SEPARATOR)
+
+        from aksetup_helper import count_down_delay
+        count_down_delay(delay=10)
