@@ -286,8 +286,8 @@ class RanluxGenerator(object):
               unsigned long idx = get_global_id(0)*4;
               while (idx + 4 < out_size)
               {
-                  *(global output_vec_t *) (output + idx) =
-                      GET_RANDOM_NUM(RANLUX_FUNC(&ranluxclstate));
+                  output_vec_t ran = GET_RANDOM_NUM(RANLUX_FUNC(&ranluxclstate));
+                  vstore4(ran, 0, &output[idx]);
                   idx += 4*NUM_WORKITEMS;
               }
 
@@ -594,8 +594,8 @@ class Random123GeneratorBase(object):
                 unsigned long idx = get_global_id(0)*4;
                 while (idx + 4 < out_size)
                 {
-                    *(global output_vec_t *) (output + idx) =
-                        GET_RANDOM_NUM(gen_bits(&k, &c));
+                    output_vec_t ran = GET_RANDOM_NUM(gen_bits(&k, &c));
+                    vstore4(ran, 0, &output[idx]);
                     idx += 4*get_global_size(0);
                 }
 
