@@ -35,7 +35,8 @@ from pyopencl.tools import (dtype_to_ctype, bitlog2,
         KernelTemplateBase, _process_code_for_macro,
         get_arg_list_scalar_arg_dtypes,
         context_dependent_memoize,
-        _NumpyTypesKeyBuilder)
+        _NumpyTypesKeyBuilder,
+        get_arg_offset_adjuster_code)
 
 import pyopencl._mymako as mako
 from pyopencl._cluda import CLUDA_PREAMBLE
@@ -148,6 +149,8 @@ void ${kernel_name}(
     %endif
     )
 {
+    ${arg_offset_adjustment}
+
     // index K in first dimension used for carry storage
     %if use_bank_conflict_avoidance:
         // Avoid bank conflicts by adding a single 32-bit value to the size of
@@ -157,16 +160,14 @@ void ${kernel_name}(
             scan_type value;
             int dummy;
         };
-        LOCAL_MEM struct wrapped_scan_type ldata[K + 1][WG_SIZE + 1];
     %else:
         struct wrapped_scan_type
         {
             scan_type value;
         };
-
-        // padded in WG_SIZE to avoid bank conflicts
-        LOCAL_MEM struct wrapped_scan_type ldata[K + 1][WG_SIZE];
     %endif
+    // padded in WG_SIZE to avoid bank conflicts
+    LOCAL_MEM struct wrapped_scan_type ldata[K + 1][WG_SIZE + 1];
 
     %if is_segmented:
         LOCAL_MEM char l_segment_start_flags[K][WG_SIZE];
@@ -329,16 +330,17 @@ void ${kernel_name}(
                 const index_type offset_end = interval_end - unit_base;
             %endif
 
-            for(index_type k = 1; k < K; k++)
+            for (index_type k = 1; k < K; k++)
             {
                 %if is_tail:
-                if (K * LID_0 + k < offset_end)
+                if ((index_type) (K * LID_0 + k) < offset_end)
                 %endif
                 {
                     scan_type tmp = ldata[k][LID_0].value;
-                    index_type seq_i = unit_base + K*LID_0 + k;
 
                     %if is_segmented:
+                    index_type seq_i = unit_base + K*LID_0 + k;
+
                     if (l_segment_start_flags[k][LID_0])
                     {
                         first_segment_start_in_k_group = min(
@@ -620,6 +622,8 @@ void ${name_prefix}_final_update(
     %endif
     )
 {
+    ${arg_offset_adjustment}
+
     %if use_lookbehind_update:
         LOCAL_MEM scan_type ldata[WG_SIZE];
     %endif
@@ -832,6 +836,8 @@ _IGNORED_WORDS = set("""
 
         a b prev_item i last_item prev_value
         N NO_SEG_BOUNDARY across_seg_boundary
+
+        arg_offset_adjustment
         """.split())
 
 
@@ -1000,7 +1006,7 @@ class _GenericScanKernelBase(object):
             resulting in a C `bool` value that determines whether a new
             scan segments starts at index *i*.  If given, makes the scan a
             segmented scan. Has access to the current index `i`, the result
-            of *input_expr* as a, and in addition may use *arguments* and
+            of *input_expr* as `a`, and in addition may use *arguments* and
             *input_fetch_expr* variables just like *input_expr*.
 
             If it returns true, then previous sums will not spill over into the
@@ -1227,8 +1233,17 @@ class GenericScanKernel(_GenericScanKernelBase):
             max_scan_wg_size = min(dev.max_work_group_size for dev in self.devices)
             wg_size_multiples = 64
 
+        # Intel beignet fails "Out of shared local memory" in test_scan int64
+        # and asserts in test_sort with this enabled:
+        # https://github.com/inducer/pyopencl/pull/238
+        # A beignet bug report (outside of pyopencl) suggests packed structs
+        # (which this is) can even give wrong results:
+        # https://bugs.freedesktop.org/show_bug.cgi?id=98717
+        # TODO: does this also affect Intel Compute Runtime?
         use_bank_conflict_avoidance = (
-                self.dtype.itemsize > 4 and self.dtype.itemsize % 8 == 0 and is_gpu)
+                self.dtype.itemsize > 4 and self.dtype.itemsize % 8 == 0
+                and is_gpu
+                and "beignet" not in self.devices[0].platform.version.lower())
 
         # k_group_size should be a power of two because of in-kernel
         # division by that number.
@@ -1339,6 +1354,7 @@ class GenericScanKernel(_GenericScanKernelBase):
         final_update_src = str(final_update_tpl.render(
             wg_size=update_wg_size,
             output_statement=self.output_statement,
+            arg_offset_adjustment=get_arg_offset_adjuster_code(self.parsed_args),
             argument_signature=", ".join(
                 arg.declarator() for arg in self.parsed_args),
             is_segment_start_expr=self.is_segment_start_expr,
@@ -1414,6 +1430,7 @@ class GenericScanKernel(_GenericScanKernelBase):
             wg_size=wg_size,
             input_expr=input_expr,
             k_group_size=k_group_size,
+            arg_offset_adjustment=get_arg_offset_adjuster_code(arguments),
             argument_signature=", ".join(arg.declarator() for arg in arguments),
             is_segment_start_expr=is_segment_start_expr,
             input_fetch_exprs=input_fetch_exprs,
@@ -1465,10 +1482,12 @@ class GenericScanKernel(_GenericScanKernelBase):
             return cl.enqueue_marker(queue, wait_for=wait_for)
 
         data_args = []
-        from pyopencl.tools import VectorArg
         for arg_descr, arg_val in zip(self.parsed_args, args):
+            from pyopencl.tools import VectorArg
             if isinstance(arg_descr, VectorArg):
-                data_args.append(arg_val.data)
+                data_args.append(arg_val.base_data)
+                if arg_descr.with_offset:
+                    data_args.append(arg_val.offset)
             else:
                 data_args.append(arg_val)
 
@@ -1576,6 +1595,8 @@ void ${name_prefix}_debug_scan(
     scan_type current = ${neutral};
     scan_type prev;
 
+    ${arg_offset_adjustment}
+
     for (index_type i = 0; i < N; ++i)
     {
         %for name, arg_name, ife_offset in input_fetch_exprs:
@@ -1629,6 +1650,7 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
         scan_tpl = _make_template(DEBUG_SCAN_TEMPLATE)
         scan_src = str(scan_tpl.render(
             output_statement=self.output_statement,
+            arg_offset_adjustment=get_arg_offset_adjuster_code(self.parsed_args),
             argument_signature=", ".join(
                 arg.declarator() for arg in self.parsed_args),
             is_segment_start_expr=self.is_segment_start_expr,
@@ -1673,7 +1695,9 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
         from pyopencl.tools import VectorArg
         for arg_descr, arg_val in zip(self.parsed_args, args):
             if isinstance(arg_descr, VectorArg):
-                data_args.append(arg_val.data)
+                data_args.append(arg_val.base_data)
+                if arg_descr.with_offset:
+                    data_args.append(arg_val.offset)
             else:
                 data_args.append(arg_val)
 
