@@ -35,7 +35,7 @@ import numpy as np
 from decorator import decorator
 import pyopencl as cl
 from pytools import memoize, memoize_method
-from pyopencl.cffi_cl import _lib
+from pyopencl._cl import bitlog2  # noqa: F401
 from pytools.persistent_dict import KeyBuilder as KeyBuilderBase
 
 import re
@@ -60,9 +60,11 @@ _register_types()
 
 # {{{ imported names
 
-bitlog2 = _lib.bitlog2
-from pyopencl.mempool import (  # noqa
-        PooledBuffer, DeferredAllocator, ImmediateAllocator, MemoryPool)
+from pyopencl._cl import (  # noqa
+        PooledBuffer as PooledBuffer,
+        _tools_DeferredAllocator as DeferredAllocator,
+        _tools_ImmediateAllocator as ImmediateAllocator,
+        MemoryPool as MemoryPool)
 
 # }}}
 
@@ -249,38 +251,38 @@ def pytest_generate_tests_for_pyopencl(metafunc):
 
     test_plat_and_dev = get_test_platforms_and_devices()
 
-    if ("device" in metafunc.funcargnames
-            or "ctx_factory" in metafunc.funcargnames
-            or "ctx_getter" in metafunc.funcargnames):
-        arg_dict = {}
+    arg_names = []
 
-        for platform, plat_devs in test_plat_and_dev:
-            if "platform" in metafunc.funcargnames:
-                arg_dict["platform"] = platform
+    for arg in ("platform", "device", "ctx_factory", "ctx_getter"):
+        if arg not in metafunc.funcargnames:
+            continue
 
-            for device in plat_devs:
-                if "device" in metafunc.funcargnames:
-                    arg_dict["device"] = device
+        if arg == "ctx_getter":
+            from warnings import warn
+            warn("The 'ctx_getter' arg is deprecated in "
+                    "favor of 'ctx_factory'.",
+                    DeprecationWarning)
 
-                if "ctx_factory" in metafunc.funcargnames:
-                    arg_dict["ctx_factory"] = ContextFactory(device)
+        arg_names.append(arg)
 
-                if "ctx_getter" in metafunc.funcargnames:
-                    from warnings import warn
-                    warn("The 'ctx_getter' arg is deprecated in "
-                            "favor of 'ctx_factory'.",
-                            DeprecationWarning)
-                    arg_dict["ctx_getter"] = ContextFactory(device)
+    arg_values = []
 
-                metafunc.addcall(funcargs=arg_dict.copy(),
-                        id=", ".join("%s=%s" % (arg, value)
-                                for arg, value in six.iteritems(arg_dict)))
+    for platform, plat_devs in test_plat_and_dev:
+        if arg_names == ["platform"]:
+            arg_values.append((platform,))
+            continue
 
-    elif "platform" in metafunc.funcargnames:
-        for platform, plat_devs in test_plat_and_dev:
-            metafunc.addcall(
-                    funcargs=dict(platform=platform),
-                    id=str(platform))
+        arg_dict = {"platform": platform}
+
+        for device in plat_devs:
+            arg_dict["device"] = device
+            arg_dict["ctx_factory"] = ContextFactory(device)
+            arg_dict["ctx_getter"] = ContextFactory(device)
+
+            arg_values.append(tuple(arg_dict[name] for name in arg_names))
+
+    if arg_names:
+        metafunc.parametrize(arg_names, arg_values, ids=str)
 
 
 # {{{ C argument lists
@@ -398,7 +400,6 @@ def get_arg_offset_adjuster_code(arg_types):
 
     return "\n".join(result)
 
-
 # }}}
 
 
@@ -413,7 +414,7 @@ def get_gl_sharing_context_properties():
     if sys.platform in ["linux", "linux2"]:
         from OpenGL import GLX
         props.append(
-            (ctx_props.GL_CONTEXT_KHR, gl_platform.GetCurrentContext()))
+            (ctx_props.GL_CONTEXT_KHR, GLX.glXGetCurrentContext()))
         props.append(
                 (ctx_props.GLX_DISPLAY_KHR,
                     GLX.glXGetCurrentDisplay()))
