@@ -48,7 +48,6 @@ else:
 def _skip_if_pocl(plat, up_to_version, msg='unsupported by pocl'):
     if plat.vendor == "The pocl project":
         if up_to_version is None or get_pocl_version(plat) <= up_to_version:
-            import pytest
             pytest.skip(msg)
 
 
@@ -122,7 +121,7 @@ def test_get_info(ctx_factory):
     def do_test(cl_obj, info_cls, func=None, try_attr_form=True):
         if func is None:
             def func(info):
-                cl_obj.get_info(info)
+                    cl_obj.get_info(info)
 
         for info_name in dir(info_cls):
             if not info_name.startswith("_") and info_name != "to_string":
@@ -182,9 +181,20 @@ def test_get_info(ctx_factory):
 
     do_test(a_buf, cl.mem_info)
 
-    kernel = prg.sum
+    kernel = prg.all_kernels()[0]
     do_test(kernel, cl.kernel_info)
 
+    for i in range(2):  # exercise cache
+        for info_name in dir(cl.kernel_work_group_info):
+            if not info_name.startswith("_") and info_name != "to_string":
+                try:
+                    print("kernel_wg_info: %s" % info_name)
+                    kernel.get_work_group_info(
+                            getattr(cl.kernel_work_group_info, info_name),
+                            device)
+                except cl.LogicError as err:
+                    print("<error: %s>" % err)
+
     evt = kernel(queue, (n,), None, a_buf)
     do_test(evt, cl.event_info)
 
@@ -295,7 +305,9 @@ def test_image_format_constructor():
 
     assert iform.channel_order == cl.channel_order.RGBA
     assert iform.channel_data_type == cl.channel_type.FLOAT
-    assert not iform.__dict__
+
+    if not cl._PYPY:
+        assert not hasattr(iform, "__dict__")
 
 
 def test_device_topology_amd_constructor():
@@ -306,7 +318,8 @@ def test_device_topology_amd_constructor():
     assert topol.device == 4
     assert topol.function == 5
 
-    assert not topol.__dict__
+    if not cl._PYPY:
+        assert not hasattr(topol, "__dict__")
 
 
 def test_nonempty_supported_image_formats(ctx_factory):
@@ -351,7 +364,7 @@ def test_that_python_args_fail(ctx_factory):
     prg.mult(queue, a.shape, None, a_buf, np.float32(2), np.int32(3))
 
     a_result = np.empty_like(a)
-    cl.enqueue_read_buffer(queue, a_buf, a_result).wait()
+    cl.enqueue_copy(queue, a_buf, a_result).wait()
 
 
 def test_image_2d(ctx_factory):
@@ -513,8 +526,8 @@ def test_copy_buffer(ctx_factory):
     buf1 = cl.Buffer(context, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
     buf2 = cl.Buffer(context, mf.WRITE_ONLY, b.nbytes)
 
-    cl.enqueue_copy_buffer(queue, buf1, buf2).wait()
-    cl.enqueue_read_buffer(queue, buf2, b).wait()
+    cl.enqueue_copy(queue, buf2, buf1).wait()
+    cl.enqueue_copy(queue, b, buf2).wait()
 
     assert la.norm(a - b) == 0
 
@@ -569,7 +582,7 @@ def test_vector_args(ctx_factory):
 
     prg.set_vec(queue, dest.shape, None, x, dest_buf)
 
-    cl.enqueue_read_buffer(queue, dest_buf, dest).wait()
+    cl.enqueue_copy(queue, dest, dest_buf).wait()
 
     assert (dest == x).all()
 
@@ -632,7 +645,13 @@ def test_enqueue_barrier_marker(ctx_factory):
     # Still relevant on pocl 1.0RC1.
     _skip_if_pocl(
             ctx.devices[0].platform, (1, 0), 'pocl crashes on enqueue_barrier')
+
     queue = cl.CommandQueue(ctx)
+
+    if queue._get_cl_version() >= (1, 2) and cl.get_cl_header_version() <= (1, 1):
+        pytest.skip("CL impl version >= 1.2, header version <= 1.1--cannot be sure "
+                "that clEnqueueWaitForEvents is implemented")
+
     cl.enqueue_barrier(queue)
     evt1 = cl.enqueue_marker(queue)
     evt2 = cl.enqueue_marker(queue, wait_for=[evt1])
@@ -648,8 +667,8 @@ def test_wait_for_events(ctx_factory):
 
 
 def test_unload_compiler(platform):
-    if (platform._get_cl_version() < (1, 2) or
-            cl.get_cl_header_version() < (1, 2)):
+    if (platform._get_cl_version() < (1, 2)
+            or cl.get_cl_header_version() < (1, 2)):
         from pytest import skip
         skip("clUnloadPlatformCompiler is only available in OpenCL 1.2")
     _skip_if_pocl(platform, (0, 13), 'pocl does not support unloading compiler')
@@ -659,36 +678,6 @@ def test_unload_compiler(platform):
     cl.unload_platform_compiler(platform)
 
 
-def test_enqueue_task(ctx_factory):
-    ctx = ctx_factory()
-    queue = cl.CommandQueue(ctx)
-    mf = cl.mem_flags
-
-    prg = cl.Program(ctx, """
-    __kernel void
-    reverse(__global const float *in, __global float *out, int n)
-    {
-        for (int i = 0;i < n;i++) {
-            out[i] = in[n - 1 - i];
-        }
-    }
-    """).build()
-    knl = prg.reverse
-
-    n = 100
-    a = np.random.rand(n).astype(np.float32)
-    b = np.empty_like(a)
-
-    buf1 = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=a)
-    buf2 = cl.Buffer(ctx, mf.WRITE_ONLY, b.nbytes)
-
-    knl.set_args(buf1, buf2, np.int32(n))
-    cl.enqueue_task(queue, knl)
-
-    cl.enqueue_copy(queue, b, buf2).wait()
-    assert la.norm(a[::-1] - b) == 0
-
-
 def test_platform_get_devices(ctx_factory):
     ctx = ctx_factory()
     platform = ctx.devices[0].platform
@@ -699,8 +688,8 @@ def test_platform_get_devices(ctx_factory):
 
     dev_types = [cl.device_type.ACCELERATOR, cl.device_type.ALL,
                  cl.device_type.CPU, cl.device_type.DEFAULT, cl.device_type.GPU]
-    if (platform._get_cl_version() >= (1, 2) and
-            cl.get_cl_header_version() >= (1, 2)
+    if (platform._get_cl_version() >= (1, 2)
+            and cl.get_cl_header_version() >= (1, 2)
             and not platform.name.lower().startswith("nvidia")):
         dev_types.append(cl.device_type.CUSTOM)
 
@@ -717,8 +706,8 @@ def test_platform_get_devices(ctx_factory):
 
 def test_user_event(ctx_factory):
     ctx = ctx_factory()
-    if (ctx._get_cl_version() < (1, 1) and
-            cl.get_cl_header_version() < (1, 1)):
+    if (ctx._get_cl_version() < (1, 1)
+            and cl.get_cl_header_version() < (1, 1)):
         from pytest import skip
         skip("UserEvent is only available in OpenCL 1.1")
 
@@ -762,6 +751,10 @@ def test_user_event(ctx_factory):
 
 
 def test_buffer_get_host_array(ctx_factory):
+    if cl._PYPY:
+        # FIXME
+        pytest.xfail("Buffer.get_host_array not yet working on pypy")
+
     ctx = ctx_factory()
     mf = cl.mem_flags
 
@@ -769,8 +762,8 @@ def test_buffer_get_host_array(ctx_factory):
     buf = cl.Buffer(ctx, mf.READ_WRITE | mf.USE_HOST_PTR, hostbuf=host_buf)
     host_buf2 = buf.get_host_array(25, np.float32)
     assert (host_buf == host_buf2).all()
-    assert (host_buf.__array_interface__['data'][0] ==
-            host_buf.__array_interface__['data'][0])
+    assert (host_buf.__array_interface__['data'][0]
+            == host_buf.__array_interface__['data'][0])
     assert host_buf2.base is buf
 
     buf = cl.Buffer(ctx, mf.READ_WRITE | mf.ALLOC_HOST_PTR, size=100)
@@ -817,7 +810,7 @@ def test_event_set_callback(ctx_factory):
     queue = cl.CommandQueue(ctx)
 
     if ctx._get_cl_version() < (1, 1):
-        pytest.skip("OpenCL 1.1 or newer required fro set_callback")
+        pytest.skip("OpenCL 1.1 or newer required for set_callback")
 
     a_np = np.random.rand(50000).astype(np.float32)
     b_np = np.random.rand(50000).astype(np.float32)
@@ -851,9 +844,17 @@ def test_event_set_callback(ctx_factory):
 
     queue.finish()
 
+    counter = 0
+
     # yuck
-    from time import sleep
-    sleep(0.1)
+    while not got_called:
+        from time import sleep
+        sleep(0.01)
+
+        # wait up to five seconds (?!)
+        counter += 1
+        if counter >= 500:
+            break
 
     assert got_called
 
@@ -886,8 +887,8 @@ def test_global_offset(ctx_factory):
 
 def test_sub_buffers(ctx_factory):
     ctx = ctx_factory()
-    if (ctx._get_cl_version() < (1, 1) or
-            cl.get_cl_header_version() < (1, 1)):
+    if (ctx._get_cl_version() < (1, 1)
+            or cl.get_cl_header_version() < (1, 1)):
         from pytest import skip
         skip("sub-buffers are only available in OpenCL 1.1")
 
@@ -916,8 +917,8 @@ def test_spirv(ctx_factory):
     ctx = ctx_factory()
     queue = cl.CommandQueue(ctx)
 
-    if (ctx._get_cl_version() < (2, 1) or
-            cl.get_cl_header_version() < (2, 1)):
+    if (ctx._get_cl_version() < (2, 1)
+            or cl.get_cl_header_version() < (2, 1)):
         from pytest import skip
         skip("SPIR-V program creation only available in OpenCL 2.1 and higher")
 
@@ -946,17 +947,10 @@ def test_coarse_grain_svm(ctx_factory):
 
     dev = ctx.devices[0]
 
-    has_svm = (ctx._get_cl_version() >= (2, 0) and
-                cl.get_cl_header_version() >= (2, 0))
-
-    if dev.platform.name == "Portable Computing Language":
-        has_svm = (
-                get_pocl_version(dev.platform) >= (1, 0)
-                and cl.get_cl_header_version() >= (2, 0))
-
-    if not has_svm:
-        from pytest import skip
-        skip("SVM only available in OpenCL 2.0 and higher")
+    from pyopencl.characterize import has_coarse_grain_buffer_svm
+    from pytest import skip
+    if not has_coarse_grain_buffer_svm(queue.device):
+        skip("device does not support coarse-grain SVM")
 
     if ("AMD" in dev.platform.name
             and dev.type & cl.device_type.CPU):
@@ -1005,13 +999,9 @@ def test_fine_grain_svm(ctx_factory):
     ctx = ctx_factory()
     queue = cl.CommandQueue(ctx)
 
+    from pyopencl.characterize import has_fine_grain_buffer_svm
     from pytest import skip
-    if (ctx._get_cl_version() < (2, 0) or
-            cl.get_cl_header_version() < (2, 0)):
-        skip("SVM only available in OpenCL 2.0 and higher")
-
-    if not (ctx.devices[0].svm_capabilities
-            & cl.device_svm_capabilities.FINE_GRAIN_BUFFER):
+    if not has_fine_grain_buffer_svm(queue.device):
         skip("device does not support fine-grain SVM")
 
     n = 3000
@@ -1043,6 +1033,10 @@ def test_fine_grain_svm(ctx_factory):
     cl.cltypes.uint2,
     ])
 def test_map_dtype(ctx_factory, dtype):
+    if cl._PYPY:
+        # FIXME
+        pytest.xfail("enqueue_map_buffer not yet working on pypy")
+
     ctx = ctx_factory()
     queue = cl.CommandQueue(ctx)
 
@@ -1058,6 +1052,35 @@ def test_map_dtype(ctx_factory, dtype):
         assert array.dtype == dt
 
 
+def test_compile_link(ctx_factory):
+    ctx = ctx_factory()
+
+    if ctx._get_cl_version() < (1, 2) or cl.get_cl_header_version() < (1, 2):
+        pytest.skip("Context and ICD loader must understand CL1.2 for compile/link")
+
+    platform = ctx.devices[0].platform
+    if platform.name == "Apple":
+        pytest.skip("Apple doesn't like our compile/link test")
+
+    queue = cl.CommandQueue(ctx)
+    vsink_prg = cl.Program(ctx, """//CL//
+        void value_sink(float x)
+        {
+        }
+        """).compile()
+    main_prg = cl.Program(ctx, """//CL//
+        void value_sink(float x);
+
+        __kernel void experiment()
+        {
+            value_sink(3.1415f + get_global_id(0));
+        }
+        """).compile()
+    z = cl.link_program(ctx, [vsink_prg, main_prg], devices=ctx.devices)
+    z.experiment(queue, (128**2,), (128,))
+    queue.finish()
+
+
 if __name__ == "__main__":
     # make sure that import failures get reported, instead of skipping the tests.
     import pyopencl  # noqa
@@ -1066,5 +1089,5 @@ if __name__ == "__main__":
     if len(sys.argv) > 1:
         exec(sys.argv[1])
     else:
-        from py.test.cmdline import main
+        from pytest import main
         main([__file__])
